from sqlparse import format  # type: ignore


def format_query(query: str) -> str:
    """
    function, that reformats the query using sqlparse.s
    """
    return format(query, reindent=True, indent_width=4)



import logging

from firebolt.common.exception import FireboltError
from firebolt.db.connection import Connection

from firebolt_ingest.aws_settings import (
    AWSSettings,
    generate_aws_credentials_string,
)
from firebolt_ingest.table_model import FILE_METADATA_COLUMNS, Table
from firebolt_ingest.table_utils import (
    does_table_exist,
    drop_table,
    execute_set_statements,
    get_table_columns,
    get_table_schema,
    verify_ingestion_file_names,
    verify_ingestion_rowcount,
)
from firebolt_ingest.utils import format_query

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s :: %(levelname)s :: %(filename)s ::\
 %(funcName)s :: %(message)s",
)

logger = logging.getLogger(__name__)


class TableService:
    def __init__(
        self,
        table: Table,
        connection: Connection,
        external_prefix: str = "ex_",
        internal_prefix: str = "",
    ):
        """
        Table service class used for creation of external/internal tables and
        performing ingestion from external into internal table

        Args:
            table (Table): An object representing the table definition. This is used for
                creating external and internal tables and for data ingestion processes.
            connection (Connection): A database connection object used to interact with
                the database or engine where the tables are managed.
            external_prefix (str, optional): A prefix string added to the table name to
                create the name of the external table. Defaults to 'ex_'.
            internal_prefix (str, optional): A prefix string added to the table name to
                create the name of the internal table. Defaults to an empty string.
        """
        self.connection = connection
        self.table = table
        self.internal_table_name = f"{internal_prefix}{self.table.table_name}"
        self.external_table_name = f"{external_prefix}{self.table.table_name}"

    def create_external_table(self, aws_settings: AWSSettings) -> None:
        """
        Constructs a query for creating an external table and executes it.

        Args:
            table: table definition
            aws_settings: aws settings
        """
        # Prepare aws credentials
        if aws_settings.aws_credentials:
            cred_stmt, cred_params = generate_aws_credentials_string(
                aws_settings.aws_credentials
            )
            cred_stmt = f"CREDENTIALS = {cred_stmt}"
        else:
            cred_stmt, cred_params = "", []

        if not self.table.s3_url and not aws_settings.s3_url:
            raise FireboltError("S3 URL wasn't provided")

        # Prepare columns
        columns_stmt, columns_params = self.table.generate_external_columns_string()

        # Generate query
        query = (
            f"CREATE EXTERNAL TABLE {self.external_table_name}\n"
            f"({columns_stmt})\n"
            f"{cred_stmt}\n"
            f"URL = ?\n"
            f"OBJECT_PATTERN = ?\n"
            f"TYPE = ({self.table.generate_file_type()})\n"
        )
        if self.table.compression:
            query += f"COMPRESSION = {self.table.compression}\n"

        params = (
            cred_params
            + columns_params
            + [self.table.s3_url if self.table.s3_url else aws_settings.s3_url]
            + [self.table.object_pattern]
        )

        logger.info(f"Create external table with query:\n{query}")
        # Execute parametrized query
        self.connection.cursor().execute(format_query(query), params)

    def create_internal_table(self, add_file_metadata=True) -> None:
        """
        Constructs a query for creating an internal table and executes it

        Args:
            table: table definition
        """

        columns_stmt, columns_params = self.table.generate_internal_columns_string(
            add_file_metadata
        )
        query = (
            f"CREATE FACT TABLE {self.internal_table_name}\n"
            f"({columns_stmt})\n"
            f"PRIMARY INDEX {self.table.generate_primary_index_string()}\n"
        )

        if self.table.partitions:
            query += f"PARTITION BY {self.table.generate_partitions_string()}\n"  # noqa: E501

        logger.info(f"Create internal table with query:\n{query}")
        self.connection.cursor().execute(format_query(query), columns_params)

    def insert_full_overwrite(
        self,
        **kwargs,
    ) -> None:
        """
        Perform a full overwrite from an external table into an internal table.
        All existing data in the internal table will be deleted, and data from the
        external table will be loaded.

        This function is appropriate for unpartitioned tables, or for
        partitioned tables you wish to fully overwrite.

        Kwargs:
            advanced_mode: (Optional)
            use_short_column_path_parquet: (Optional) Use short parquet column path
             and skipping repeated nodes and their child node
        """
        cursor = self.connection.cursor()

        # TODO: uncomment it back after the fix applied,
        # commented because of https://packboard.atlassian.net/browse/FIR-26886
        # raise_on_tables_non_compatibility(cursor,
        #                                   self.table,
        #                                   ignore_meta_columns=True)

        # get table schema
        internal_table_schema = get_table_schema(cursor, self.internal_table_name)
        internal_table_columns = get_table_columns(cursor, self.internal_table_name)

        # drop the table
        logger.info(f"Drop internal table: {self.internal_table_name}")
        drop_table(cursor, self.internal_table_name)

        # recreate the table
        logger.info(f"Create internal table:\n{internal_table_schema}")
        cursor.execute(query=internal_table_schema)

        # insert the data from external to internal
        column_names = [
            (f'"{c.name}"' + (f" AS {c.alias}" if c.alias else ""))
            for c in self.table.columns
        ]

        for c in FILE_METADATA_COLUMNS:
            name, type_ = c.name, c.type
            # Check if FILE_METADATA_COLUMNS is present in internal_table_columns
            # TIMESTAMPNTZ is sometimes represented as TIMESTAMP, need to check both
            if (name, type_) in internal_table_columns or (
                type_ == "TIMESTAMPNTZ"
                and (name, "TIMESTAMP") in internal_table_columns
            ):
                column_names.append(name)

        insert_query = (
            f"INSERT INTO {self.internal_table_name}\n"
            f"SELECT {', '.join(column_names)}\n"
            f"FROM {self.external_table_name}\n"
        )

        logger.info(f"Insert with query:\n{insert_query}")
        execute_set_statements(
            cursor,
            **kwargs,
        )
        cursor.execute(query=format_query(insert_query))

    def insert_incremental_append(self, use_materialized_query=False, **kwargs) -> None:
        """
        Insert from the external table only new files,
        that aren't in the internal table.

        Requires internal table to have file-metadata columns
        (source_file_name and source_file_timestamp)

        Args:
        use_materialized_query (bool):
            If set to True, the function uses an materialized query
                strategy for insertion.
            If set to False (default), the function uses the standard query approach.
        **kwargs: Additional keyword arguments which may be passed to other functions
            used in this method.
        Returns:

        """
        cursor = self.connection.cursor()
        # TODO: uncomment it back after the fix applied,
        # commented because of https://packboard.atlassian.net/browse/FIR-26886
        # raise_on_tables_non_compatibility(cursor,
        #                                   self.table,
        #                                   ignore_meta_columns=False)

        if not does_table_exist(cursor, self.internal_table_name):
            raise FireboltError(f"Fact table {self.internal_table_name} doesn't exist")
        if not does_table_exist(cursor, self.external_table_name):
            raise FireboltError(
                f"External table {self.external_table_name} doesn't exist"
            )

        column_names = [
            (f'"{c.name}"' + (f" AS {c.alias}" if c.alias else ""))
            for c in self.table.columns
        ]

        if use_materialized_query:
            # Optimized query
            insert_query = f"""
                INSERT INTO {self.internal_table_name}
                WITH a AS materialized (
                    SELECT DISTINCT source_file_name
                    FROM {self.internal_table_name}
                )
                SELECT {', '.join(column_names)},
                    source_file_name, source_file_timestamp
                FROM {self.external_table_name}
                WHERE source_file_name NOT IN (
                    SELECT source_file_name
                    FROM a
                )
            """
        else:
            insert_query = f"""
                INSERT INTO {self.internal_table_name}
                SELECT {', '.join(column_names)},
                        source_file_name, source_file_timestamp
                FROM {self.external_table_name}
                WHERE (source_file_name, source_file_timestamp::timestampntz)
                NOT IN (
                    SELECT DISTINCT source_file_name,
                                    source_file_timestamp
                    FROM {self.internal_table_name})
                """

        logger.info(f"Insert with query:\n{insert_query}")
        execute_set_statements(
            cursor,
            **kwargs,
        )
        cursor.execute(query=format_query(insert_query))

    def verify_ingestion(self) -> bool:
        """
        verify ingestion by running a sequence of verification, currently implemented:
        - verification by rowcount
        """

        cursor = self.connection.cursor()
        return verify_ingestion_rowcount(
            cursor, self.internal_table_name, self.external_table_name
        ) and verify_ingestion_file_names(cursor, self.internal_table_name)

    def insert(self, use_materialized_query=False, **kwargs) -> None:
        """
        Inserts data into a table based on the synchronization mode specified
        in the table's configuration.

        This method checks the `sync_mode` attribute of the associated table
        and performs an insert operation accordingly.
        If the `sync_mode` is set to "overwrite", it performs
        a full overwrite of the data in the table.
        If it's set to "append", it appends the new data incrementally.
        For any other `sync_mode` values, a ValueError is raised, indicating
        an uncertain sync mode configuration.

        Args:
        use_materialized_query (bool):
            If set to True, the function uses an materialized query
                strategy for insertion.
            If set to False (default), the function uses the standard query approach.
        **kwargs: Additional keyword arguments which may be passed to other functions
            used in this method.
        """
        if self.table.sync_mode == "overwrite":
            self.insert_full_overwrite(**kwargs)
        elif self.table.sync_mode == "append":
            self.insert_incremental_append(
                use_materialized_query=use_materialized_query, **kwargs
            )
        else:
            raise ValueError(
                "Uncertain sync mode in config \
                use insert_full_overwrite/insert_incremental_append instead"
            )

    def drop_internal_table(self) -> None:
        """
        Drops the internal table associated with the current object.
        """
        logger.info(f"Drop internal table: {self.internal_table_name}")
        cursor = self.connection.cursor()
        drop_table(cursor, self.internal_table_name)

    def drop_external_table(self) -> None:
        """
        Drops the external table associated with the current object.
        """
        logger.info(f"Drop external table: {self.external_table_name}")
        cursor = self.connection.cursor()
        drop_table(cursor, self.external_table_name)

    def drop_tables(self) -> None:
        """
        Drops both internal and external tables associated with the current object.
        """
        self.drop_internal_table()
        self.drop_external_table()

    def does_external_table_exist(self) -> bool:
        """
        Checks if the external table exists in the database.
        """
        return does_table_exist(self.connection.cursor(), self.external_table_name)

    def does_internal_table_exist(self) -> bool:
        """
        Checks if the internal table exists in the database.
        """
        return does_table_exist(self.connection.cursor(), self.internal_table_name)



from typing import List, Optional, Tuple

from pydantic import BaseSettings, Field, SecretStr, root_validator


class AWSCredentialsKeySecret(BaseSettings):
    aws_key_id: str
    aws_secret_key: SecretStr


class AWSCredentialsRole(BaseSettings):
    role_arn: SecretStr
    external_id: Optional[str]


class AWSCredentials(BaseSettings):
    key_secret_creds: Optional[AWSCredentialsKeySecret]
    # or
    role_creds: Optional[AWSCredentialsRole]

    @root_validator
    def mutual_exclusive_with_creds(cls, values: dict) -> dict:
        if values["key_secret_creds"] and values["role_creds"]:
            raise ValueError(
                "Only one should be provided: key_secret_creds or role_arn"
            )

        if values["key_secret_creds"] is None and values["role_creds"] is None:
            raise ValueError(
                "At least one should be provided: key_secret_creds or role_arn"
            )

        return values


class AWSSettings(BaseSettings):
    s3_url: Optional[str] = Field(
        min_length=1,
        max_length=255,
        regex=r"^s3:\/\/[a-z0-9-]{1,64}\/[a-zA-Z0-9-_.\/]*",
    )
    aws_credentials: Optional[AWSCredentials]


def generate_aws_credentials_string(creds: AWSCredentials) -> Tuple[str, List[str]]:
    """
        Prepares sql statement for passing the AWS credentials

    Args:
        creds: valid aws credentials

    Returns:
        a part of sql with the placeholders for the relevant credentials
        sequence of credentials
    """

    if creds.key_secret_creds is not None:
        ks_creds = creds.key_secret_creds
        return (
            "(AWS_KEY_ID = ? AWS_SECRET_KEY = ?)",
            [ks_creds.aws_key_id, ks_creds.aws_secret_key.get_secret_value()],
        )

    if creds.role_creds is not None:
        role_creds = creds.role_creds
        stmt = "(AWS_ROLE_ARN = ?{})".format(
            " AWS_ROLE_EXTERNAL_ID = ?" if role_creds.external_id else ""
        )
        params = [role_creds.role_arn.get_secret_value()] + (
            [role_creds.external_id] if role_creds.external_id else []
        )

        return stmt, params

    raise ValueError(
        "aws credentials should either have "
        "key_secret pair or role_creds; none is found"
    )



from functools import wraps
from typing import List, Sequence, Set, Tuple

from firebolt.common.exception import FireboltError
from firebolt.db import Cursor

from firebolt_ingest.table_model import FILE_METADATA_COLUMNS, Table
from firebolt_ingest.utils import format_query


def table_must_exist(func):
    @wraps(func)
    def with_table_existence_check(cursor: Cursor, table_name: str, **kwargs):
        if not does_table_exist(cursor=cursor, table_name=table_name):
            raise FireboltError(
                f"Table {table_name} does not exist when calling {func.__name__}"
            )
        return func(cursor=cursor, table_name=table_name, **kwargs)

    return with_table_existence_check


def get_table_schema(cursor: Cursor, table_name: str) -> str:
    """
    Return the create command of the existing table.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table

    Returns:
        CREATE TABLE ... command
    """
    cursor.execute("SHOW TABLES")
    data = cursor.fetchall()

    def find_column_index(columns: Sequence, column_name: str) -> int:
        result = [
            idx for idx, column in enumerate(columns) if column.name == column_name
        ]

        if len(result) == 0:
            raise FireboltError(f"Cannot find expected column: {column_name}")
        elif len(result) > 1:
            raise FireboltError(f"Too many columns: {column_name} found")

        return result[0]

    table_name_index = find_column_index(cursor.description, "table_name")
    schema_index = find_column_index(cursor.description, "schema")

    internal_table_schema = [
        row[schema_index]
        for row in data  # type: ignore
        if row[table_name_index] == table_name
    ]

    if len(internal_table_schema) == 0:
        raise FireboltError("internal table doesn't exist")

    return str(internal_table_schema[0])


def drop_table(cursor: Cursor, table_name: str) -> None:
    """
    Drop a table.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table to drop.

    Raises an exception if the table did not drop.
    """
    drop_query = f"DROP TABLE IF EXISTS {table_name} CASCADE"

    # drop the table
    cursor.execute(query=format_query(drop_query))

    # verify that the drop succeeded
    if does_table_exist(cursor, table_name):
        raise FireboltError(f"Table {table_name} did not drop successfully.")


@table_must_exist
def get_table_columns(cursor: Cursor, table_name: str) -> List[Tuple]:
    """
    Get the column names of an existing table on Firebolt.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table
    """

    cursor.execute(
        "SELECT column_name, data_type "
        "FROM information_schema.columns "
        "WHERE table_name = ?",
        [table_name],
    )
    return [(column_name, data_type) for column_name, data_type in cursor.fetchall()]


def verify_ingestion_rowcount(
    cursor: Cursor, internal_table_name: str, external_table_name: str
) -> bool:
    """
    Verify, that the fact and external table have the same number of rows

    Note: doesn't check for existence of the fact and external tables,
    hence not safe for external usage. Could lead to sql-injection

    Args:
        cursor: Firebolt database cursor
        internal_table_name: name of the fact table
        external_table_name: name of the external table

    Returns: true if the number of rows the same
    """
    query = f"""
    SELECT
        (SELECT count(*) FROM {internal_table_name}) AS rc_fact,
        (SELECT count(*) FROM {external_table_name}) AS rc_external
    """
    cursor.execute(query=format_query(query))

    data = cursor.fetchall()
    if data is None:
        return False

    return data[0][0] == data[0][1]  # type: ignore


def verify_ingestion_file_names(cursor: Cursor, internal_table_name: str) -> bool:
    """
    Verify ingestion using the metadata. If we have entries with the same
    source_file_name and different source_file_timestamp we might have duplicates
    """

    table_columns = get_table_columns(cursor, internal_table_name)

    # if the metadata is missing return True,
    # since it is not possible to do the validation
    if not {(column.name, column.type) for column in FILE_METADATA_COLUMNS}.issubset(
        table_columns
    ):
        return True

    query = f"""
    SELECT source_file_name FROM {internal_table_name}
    GROUP BY source_file_name
    HAVING count(DISTINCT source_file_timestamp) <> 1
    """
    cursor.execute(query=format_query(query))

    # if the table is correct, the number of fetched rows should be zero
    return len(cursor.fetchall()) == 0  # type: ignore


def does_table_exist(cursor: Cursor, table_name: str) -> bool:
    """
    Check whether table with table_name exists,
    and return True if it exists, False otherwise.
    """
    find_query = "SELECT * FROM information_schema.tables WHERE table_name = ?"

    return cursor.execute(find_query, [table_name]) != 0


def check_table_compatibility(
    cursor: Cursor,
    table_name: str,
    expected_table_columns: Set[Tuple[str, str]],
    skip_columns: Set[Tuple[str, str]],
) -> List[Tuple[str, str, str, str]]:
    """
    Check whether actual table columns are equivalent to expected

    Args:
        cursor: cursor for query execution
        table_name: name of the existing table, from which
         the actual columns will be extracted
        expected_table_columns: columns, that should be there
        skip_columns: all errors with the specified columns will be ignored

    Returns: a list of all incompatibilities as a list of tuple
             (column_name, column_type, table_name, message)

    """
    actual_table_columns = set(get_table_columns(cursor, table_name))

    error_list = []
    for column in actual_table_columns - expected_table_columns - skip_columns:
        error_list.append((column[0], column[1], table_name, "expected but not found"))
    for column in expected_table_columns - actual_table_columns - skip_columns:
        error_list.append((column[0], column[1], table_name, "found but not expected"))

    return error_list


def raise_on_tables_non_compatibility(
    cursor: Cursor,
    table: Table,
    ignore_meta_columns: bool,
):
    """
    Check whether internal and external tables are compatible,
    and if not raise an exception with an appropriate error message
    """
    expected_internal_columns = set(
        ((c.alias if c.alias else c.name), c.type) for c in table.columns
    )
    expected_external_columns = set((c.name, c.type) for c in table.columns)

    skip_columns = (
        [(c.name, c.type) for c in FILE_METADATA_COLUMNS] if ignore_meta_columns else []
    )
    if not ignore_meta_columns:
        expected_internal_columns.update(
            (c.name, c.type) for c in FILE_METADATA_COLUMNS
        )

    error_list = check_table_compatibility(
        cursor, table.table_name, expected_internal_columns, set(skip_columns)
    ) + check_table_compatibility(
        cursor, f"ex_{table.table_name}", expected_external_columns, set()
    )

    if error_list:
        raise FireboltError(
            "\n".join(
                f"Column ({err[0]}, {err[1]}) in table ({err[2]}) {err[3]}"
                for err in error_list
            )
        )


def execute_set_statements(cursor, **kwargs):
    """
    Execute set statements on the cursor using keyword arguments.

    Allowed kwargs:
        - advanced_mode
        - use_short_column_path_parquet
        - use_classic_parquet (only if advanced_mode==True)
        - mask_internal_errors (only if advanced_mode==True)

    Args:
        cursor: The database cursor.
        kwargs: Keyword arguments for various settings.
    """

    defaults = {
        "advanced_mode": False,
        "use_short_column_path_parquet": False,
        "use_classic_parquet": False,
        "mask_internal_errors": True,
    }

    # Update defaults with provided kwargs
    params = {**defaults, **kwargs}

    for key, value in params.items():
        if not isinstance(value, bool):
            raise ValueError(
                f"The value for '{key}' must be a boolean, got {type(value)}"
            )

    always_execute_keys = ["advanced_mode", "use_short_column_path_parquet"]
    for key in always_execute_keys:
        cursor.execute(f"set {key}={int(params[key])}")

    # Those params are allowed only if 'advanced_mode' is True
    if params["advanced_mode"]:
        for key in ["use_classic_parquet", "mask_internal_errors"]:
            cursor.execute(f"set {key}={int(params[key])}")
